from sqlalchemy import Column, \
                       Integer, \
                       String, \
                       ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
from .royals import Royal
from .mmevents import MMEvent


class MMResponse:
    __tablename__ = "mmresponse"

    @declared_attr
    def royal_id(self):
        return Column(Integer, ForeignKey("royals.uid"), primary_key=True)

    @declared_attr
    def royal(self):
        return relationship("Royal", backref="mmresponses_given")

    @declared_attr
    def mmevent_id(self):
        return Column(Integer, ForeignKey("mmevents.mmid"), primary_key=True)

    @declared_attr
    def mmevent(self):
        return relationship("MMEvent", backref="responses")

    @declared_attr
    def response(self):
        # Valid decisions are YES, LATER or NO
        return Column(String)

    def __repr__(self):
        return f"<MMResponse of {self.royal}: {self.response}>"



"""A :py:class:`royalnet.web.Royalprint` to create new Royals."""
import flask as f
import os
import bcrypt
from ...royalprint import Royalprint
from ...shortcuts import error
from ....database.tables import Royal, Alias


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("newaccount", __name__, url_prefix="/newaccount", required_tables={Royal, Alias},
                template_folder=tmpl_dir)


@rp.route("/", methods=["GET", "POST"])
def login_index():
    if f.request.method == "GET":
        return f.render_template("newaccount_index.html")
    elif f.request.method == "POST":
        alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
        fd = f.request.form
        if "username" not in fd:
            return error(400, "Non è stato inserito nessun username.")
        if "password" not in fd:
            return error(400, "Non è stata inserita nessuna password.")
        royal = alchemy_session.query(alchemy.Royal).filter_by(username=fd["username"]).one_or_none()
        if royal is not None:
            return error(403, "Esiste già un utente con quell'username.")
        alias = alchemy_session.query(alchemy.Alias).filter_by(alias=fd["username"]).one_or_none()
        if alias is not None:
            return error(403, "Esiste già un utente con quell'alias.")
        royal = alchemy.Royal(username=fd["username"],
                              password=bcrypt.hashpw(bytes(fd["password"], encoding="utf8"), bcrypt.gensalt()),
                              role="Guest")
        alchemy_session.add(royal)
        alias = alchemy.Alias(royal=royal, alias=royal.username.lower())
        alchemy_session.add(alias)
        alchemy_session.commit()
        return f.redirect(f.url_for("login.login_index"))



import re
import markdown2


class RenderError(Exception):
    """An error occurred while trying to render the page."""


def prepare_page_markdown(markdown):
    if list(markdown).count(">") > 99:
        raise RenderError("Too many nested quotes")
    converted_md = markdown2.markdown(markdown.replace("<", "&lt;"),
                                      extras=["spoiler", "tables", "smarty-pants", "fenced-code-blocks"])
    converted_md = re.sub(r"{https?://(?:www\.)?(?:youtube\.com/watch\?.*?&?v=|youtu.be/)([0-9A-Za-z-]+).*?}",
                          r'<div class="youtube-embed">'
                          r'   <iframe src="https://www.youtube-nocookie.com/embed/\1?rel=0&amp;showinfo=0"'
                          r'           frameborder="0"'
                          r'           allow="autoplay; encrypted-media"'
                          r'           allowfullscreen'
                          r'           width="640px"'
                          r'           height="320px">'
                          r'   </iframe>'
                          r'</div>', converted_md)
    converted_md = re.sub(r"{https?://clyp.it/([a-z0-9]+)}",
                          r'<div class="clyp-embed">'
                          r'    <iframe width="100%" height="160" src="https://clyp.it/\1/widget" frameborder="0">'
                          r'    </iframe>'
                          r'</div>', converted_md)
    return converted_md



import datetime
import dateparser
import os
import telegram
import asyncio
import re
import logging
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...database.tables import MMEvent, MMDecision, MMResponse
from ...error import *
from ...utils import asyncify, telegram_escape, sleep_until

log = logging.getLogger(__name__)


class MmCommand(Command):
    """Matchmaking command.

    Requires the MM_CHANNEL_ID envvar to be set."""
    name: str = "mm"

    description: str = "Trova giocatori per una partita a qualcosa."

    syntax: str = "[ (data) ] (nomegioco)\n[descrizione]"

    require_alchemy_tables = {MMEvent, MMDecision, MMResponse}

    _cycle_duration = 5

    @staticmethod
    def _main_keyboard(mmevent: MMEvent) -> typing.Optional[telegram.InlineKeyboardMarkup]:
        if mmevent.state == "WAITING":
            return telegram.InlineKeyboardMarkup([
                [telegram.InlineKeyboardButton("🔵 Ci sarò!", callback_data=f"mm_{mmevent.mmid}_d_YES")],
                [telegram.InlineKeyboardButton("⚫️ Forse...", callback_data=f"mm_{mmevent.mmid}_d_MAYBE")],
                [telegram.InlineKeyboardButton("🔴 Non mi interessa.", callback_data=f"mm_{mmevent.mmid}_d_NO")]
            ])
        elif mmevent.state == "DECISION":
            return telegram.InlineKeyboardMarkup([
                [telegram.InlineKeyboardButton("🔵 Ci sarò!", callback_data=f"mm_{mmevent.mmid}_d_YES"),
                 telegram.InlineKeyboardButton("🔴 Non mi interessa...", callback_data=f"mm_{mmevent.mmid}_d_NO")]
            ])
        elif mmevent.state == "READY_CHECK":
            return telegram.InlineKeyboardMarkup([
                [telegram.InlineKeyboardButton("🚩 Avvia la partita", callback_data=f"mm_{mmevent.mmid}_start")]
            ])
        elif mmevent.state == "STARTED":
            return None
        else:
            raise ValueError(f"state is of an unknown value ({mmevent.state})")

    @staticmethod
    def _main_text(mmevent: MMEvent):
        text = f"🌐 [{mmevent.datetime.strftime('%Y-%m-%d %H:%M')}] [b]{mmevent.title}[/b]\n"
        if mmevent.description:
            text += f"{mmevent.description}\n"
        text += "\n"
        if mmevent.state == "WAITING" or mmevent.state == "DECISION":
            for mmdecision in sorted(mmevent.decisions, key=lambda mmd: mmd.royal.username):
                mmdecision: "MMDecision"
                if mmdecision.decision == "YES":
                    text += "🔵 "
                elif mmdecision.decision == "MAYBE":
                    text += "⚫️ "
                elif mmdecision.decision == "NO":
                    text += "🔴 "
                else:
                    raise ValueError(f"decision is of an unknown value ({mmdecision.decision})")
                text += f"{mmdecision.royal}\n"
        elif mmevent.state == "READY_CHECK":
            for mmresponse in sorted(mmevent.responses, key=lambda mmr: mmr.royal.username):
                mmresponse: "MMResponse"
                if mmresponse.response is None:
                    text += "❔ "
                elif mmresponse.response == "YES":
                    text += "✅ "
                elif mmresponse.response == "LATER":
                    text += "🕒 "
                elif mmresponse.response == "NO":
                    text += "❌ "
                else:
                    raise ValueError(f"response is of an unknown value ({mmresponse.response})")
                text += f"{mmresponse.royal}\n"
        elif mmevent.state == "STARTED":
            for mmresponse in sorted(mmevent.responses, key=lambda mmr: mmr.response, reverse=True):
                if mmresponse.response == "YES":
                    text += f"✅ {mmresponse.royal}\n"
                elif mmresponse.response == "NO":
                    text += f"❌ {mmresponse.royal}\n"
        return text

    async def _run_mm(self, mmevent: MMEvent) -> None:
        client: telegram.Bot = self.interface.bot.client

        async def update_message() -> None:
            try:
                await self.interface.bot.safe_api_call(client.edit_message_text,
                                                       text=telegram_escape(self._main_text(mmevent)),
                                                       chat_id=os.environ["MM_CHANNEL_ID"],
                                                       message_id=mmevent.message_id,
                                                       parse_mode="HTML",
                                                       disable_web_page_preview=True,
                                                       reply_markup=self._main_keyboard(mmevent))
            except telegram.error.BadRequest:
                pass

        decision_string = f"⚫️ Hai detto che forse parteciperai a [b]{mmevent.title}[/b]" \
                          f" alle [b]{mmevent.datetime.strftime('%H:%M')}[/b].\n" \
                          f"Confermi di volerci essere? (Metti sì anche se arrivi un po' in ritardo!)"

        decision_keyboard = telegram.InlineKeyboardMarkup([
                [telegram.InlineKeyboardButton("🔵 Ci sarò!", callback_data=f"mm_{mmevent.mmid}_d_YES"),
                 telegram.InlineKeyboardButton("🔴 Non mi interessa più.", callback_data=f"mm_{mmevent.mmid}_d_NO")]
            ])

        async def decision_yes(data: CommandData):
            royal = await data.get_author()
            mmdecision: MMDecision = await asyncify(
                self.interface.session.query(self.interface.alchemy.MMDecision).filter_by(mmevent=mmevent,
                                                                                          royal=royal).one_or_none)
            if mmdecision is None:
                mmdecision: MMDecision = self.interface.alchemy.MMDecision(royal=royal,
                                                                           mmevent=mmevent,
                                                                           decision="YES")
                self.interface.session.add(mmdecision)
            else:
                mmdecision.decision = "YES"
            await asyncify(self.interface.session.commit)
            await update_message()
            return "🔵 Hai detto che ci sarai!"

        async def decision_maybe(data: CommandData):
            royal = await data.get_author()
            mmdecision: MMDecision = await asyncify(
                self.interface.session.query(self.interface.alchemy.MMDecision).filter_by(mmevent=mmevent,
                                                                                          royal=royal).one_or_none)
            if mmdecision is None:
                mmdecision: MMDecision = self.interface.alchemy.MMDecision(royal=royal,
                                                                           mmevent=mmevent,
                                                                           decision="MAYBE")
                self.interface.session.add(mmdecision)
            else:
                mmdecision.decision = "MAYBE"
            # Can't asyncify this
            self.interface.session.commit()
            await update_message()
            return "⚫️ Hai detto che forse ci sarai. Rispondi al messaggio di conferma 5 minuti prima dell'inizio!"

        async def decision_no(data: CommandData):
            royal = await data.get_author()
            mmdecision: MMDecision = await asyncify(
                self.interface.session.query(self.interface.alchemy.MMDecision).filter_by(mmevent=mmevent,
                                                                                          royal=royal).one_or_none)
            if mmdecision is None:
                mmdecision: MMDecision = self.interface.alchemy.MMDecision(royal=royal,
                                                                           mmevent=mmevent,
                                                                           decision="NO")
                self.interface.session.add(mmdecision)
            else:
                mmdecision.decision = "NO"
            # Can't asyncify this
            self.interface.session.commit()
            await update_message()
            return "🔴 Hai detto che non ti interessa."

        def response_string() -> str:
            delay = (datetime.datetime.now() - mmevent.datetime).total_seconds()
            if delay < 60:
                return f"🚩 E' ora di [b]{mmevent.title}[/b]!\n" \
                       f"Sei pronto?"
            return f"🕒 Sei in ritardo di [b]{int(delay / 60)} minuti[/b] per [b]{mmevent.title}[/b]...\n" \
                   f"Sei pronto?"

        response_keyboard = telegram.InlineKeyboardMarkup([
                [telegram.InlineKeyboardButton("✅ Ci sono!",
                                               callback_data=f"mm_{mmevent.mmid}_r_YES")],
                [telegram.InlineKeyboardButton("🕒 Aspettatemi ancora un po'!",
                                               callback_data=f"mm_{mmevent.mmid}_r_LATER")],
                [telegram.InlineKeyboardButton("❌ Non vengo più, mi spiace.",
                                               callback_data=f"mm_{mmevent.mmid}_r_NO")]
            ])

        async def response_yes(data: CommandData):
            royal = await data.get_author()
            mmresponse: MMResponse = await asyncify(
                self.interface.session.query(self.interface.alchemy.MMResponse).filter_by(mmevent=mmevent,
                                                                                          royal=royal).one_or_none)
            mmresponse.response = "YES"
            # Can't asyncify this
            self.interface.session.commit()
            await update_message()
            return "✅ Sei pronto!"

        async def response_later(data: CommandData):
            royal = await data.get_author()
            mmresponse: MMResponse = await asyncify(
                self.interface.session.query(self.interface.alchemy.MMResponse).filter_by(mmevent=mmevent,
                                                                                          royal=royal).one_or_none)
            mmresponse.response = "LATER"
            # Can't asyncify this
            self.interface.session.commit()
            await update_message()
            return f"🕒 Hai chiesto agli altri di aspettarti {self._cycle_duration} minuti."

        async def response_no(data: CommandData):
            royal = await data.get_author()
            mmresponse: MMResponse = await asyncify(
                self.interface.session.query(self.interface.alchemy.MMResponse).filter_by(mmevent=mmevent,
                                                                                          royal=royal).one_or_none)
            mmresponse.response = "NO"
            # Can't asyncify this
            self.interface.session.commit()
            await update_message()
            return "❌ Hai detto che non ci sarai."

        def started_string():
            text = f"🚩 L'evento [b]{mmevent.title}[/b] è iniziato!\n\n" \
                   f"Partecipano:\n"
            for mmresponse in sorted(mmevent.responses, key=lambda mmr: mmr.response, reverse=True):
                if mmresponse.response == "YES":
                    text += f"✅ {mmresponse.royal}\n"
                elif mmresponse.response == "NO":
                    text += "❌ {mmresponse.royal}\n"
            return text

        started_without_you_string = f"🚩 Non hai confermato la tua presenza in tempo e [b]{mmevent.title}[/b] è" \
                                     f" iniziato senza di te.\n" \
                                     f"Mi dispiace!"

        async def start_event():
            mmevent.state = "STARTED"
            for mmresponse in mmevent.responses:
                if mmresponse.response is None:
                    mmresponse.response = "NO"
                if mmresponse.response == "LATER":
                    mmresponse.response = "NO"

                if mmresponse.response == "YES":
                    await self.interface.bot.safe_api_call(client.send_message,
                                                           chat_id=mmresponse.royal.telegram[0].tg_id,
                                                           text=telegram_escape(started_string()),
                                                           parse_mode="HTML",
                                                           disable_webpage_preview=True)
                else:
                    await self.interface.bot.safe_api_call(client.send_message,
                                                           chat_id=mmresponse.royal.telegram[0].tg_id,
                                                           text=telegram_escape(started_without_you_string),
                                                           parse_mode="HTML",
                                                           disable_webpage_preview=True)
            await asyncify(self.interface.session.commit)
            await update_message()

        async def start_key(data: CommandData):
            royal = await data.get_author()
            if royal == mmevent.creator:
                await start_event()

        if mmevent.state == "WAITING":
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_d_YES", decision_yes)
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_d_MAYBE", decision_maybe)
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_d_NO", decision_no)
            await sleep_until(mmevent.datetime - datetime.timedelta(minutes=10))
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_d_YES")
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_d_MAYBE")
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_d_NO")
            mmevent.state = "DECISION"
            for mmdecision in mmevent.decisions:
                mmdecision: MMDecision
                if mmdecision.decision == "MAYBE":
                    await self.interface.bot.safe_api_call(client.send_message,
                                                           chat_id=mmdecision.royal.telegram[0].tg_id,
                                                           text=telegram_escape(decision_string),
                                                           parse_mode="HTML",
                                                           disable_webpage_preview=True,
                                                           reply_markup=decision_keyboard)
            await asyncify(self.interface.session.commit)
            await update_message()

        if mmevent.state == "DECISION":
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_d_YES", decision_yes)
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_d_NO", decision_no)
            await sleep_until(mmevent.datetime)
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_d_YES")
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_d_NO")
            mmevent.state = "READY_CHECK"
            for mmdecision in mmevent.decisions:
                if mmdecision.decision == "MAYBE":
                    mmdecision.decision = "NO"
                elif mmdecision.decision == "YES":
                    mmresponse: MMResponse = self.interface.alchemy.MMResponse(royal=mmdecision.royal, mmevent=mmevent)
                    self.interface.session.add(mmresponse)
            await asyncify(self.interface.session.commit)
            await update_message()

        if mmevent.state == "READY_CHECK":
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_r_YES", response_yes)
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_r_LATER", response_later)
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_r_NO", response_no)
            self.interface.register_keyboard_key(f"mm_{mmevent.mmid}_forcestart", start_key)
            cycle = 0
            while True:
                for mmresponse in mmevent.responses:
                    # Send messages
                    if mmresponse.response is None:
                        await self.interface.bot.safe_api_call(client.send_message,
                                                               chat_id=mmresponse.royal.telegram[0].tg_id,
                                                               text=telegram_escape(response_string()),
                                                               parse_mode="HTML",
                                                               disable_webpage_preview=True,
                                                               reply_markup=response_keyboard)
                # Wait
                await asyncio.sleep(60 * self._cycle_duration)
                # Advance cycle
                for mmresponse in mmevent.responses:
                    if mmresponse.response is None:
                        mmresponse.response = "NO"
                    if mmresponse.response == "LATER":
                        mmresponse.response = None
                # Check if the event can start
                for mmresponse in mmevent.responses:
                    if mmresponse.response is None:
                        break
                else:
                    break
                cycle += 1

            await start_event()
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_r_YES")
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_r_LATER")
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_r_NO")
            self.interface.unregister_keyboard_key(f"mm_{mmevent.mmid}_forcestart")

    def __init__(self, interface):
        super().__init__(interface)
        if self.interface.name != "telegram":
            return
        log.debug("Loading pending MMEvents from the database")
        mmevents = self.interface.session.query(self.interface.alchemy.MMEvent) \
                                         .filter(self.interface.alchemy.MMEvent.datetime > datetime.datetime.now()) \
                                         .all()
        log.info(f"Found {len(mmevents)} pending MMEvents")
        for mmevent in mmevents:
            interface.loop.create_task(self._run_mm(mmevent))

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self.interface.name != "telegram":
            raise UnsupportedError("mm is supported only on Telegram")
        client: telegram.Bot = self.interface.bot.client
        creator = await data.get_author(error_if_none=True)
        try:
            timestring, title, description = args.match(r"\[\s*([^]]+)\s*]\s*([^\n]+)\s*\n?\s*(.+)?\s*", re.DOTALL)
        except InvalidInputError:
            timestring, title, description = args.match(r"\s*(.+?)\s*\n\s*([^\n]+)\s*\n?\s*(.+)?\s*", re.DOTALL)
        try:
            dt: typing.Optional[datetime.datetime] = dateparser.parse(timestring)
        except OverflowError:
            dt = None
        if dt is None:
            await data.reply("⚠️ La data che hai specificato non è valida.")
            return
        if dt <= datetime.datetime.now():
            await data.reply("⚠️ La data che hai specificato è nel passato.")
            return
        mmevent: MMEvent = self.interface.alchemy.MMEvent(creator=creator,
                                                          datetime=dt,
                                                          title=title,
                                                          description=description,
                                                          state="WAITING")
        self.interface.session.add(mmevent)
        await asyncify(self.interface.session.commit)

        message: telegram.Message = await self.interface.bot.safe_api_call(client.send_message,
                                                                           chat_id=os.environ["MM_CHANNEL_ID"],
                                                                           text=telegram_escape(self._main_text(mmevent)),
                                                                           parse_mode="HTML",
                                                                           disable_webpage_preview=True,
                                                                           reply_markup=self._main_keyboard(mmevent))

        mmevent.message_id = message.message_id
        # Can't asyncify this
        await asyncify(self.interface.session.commit)

        await self._run_mm(mmevent)



import typing
import aiohttp
import sortedcontainers
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ..commandinterface import CommandInterface
from ...utils import parse_5etools_entry, ordinalformat, andformat


class DndspellCommand(Command):
    name: str = "dndspell"

    description: str = "Ottieni informazioni su una magia di D&D5e."

    syntax = "(nomemagia)"

    _dnddata: sortedcontainers.SortedKeyList = None

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.loop.create_task(self._fetch_dnddata())

    async def _fetch_dnddata(self):
        self._dnddata = self._dnddata = sortedcontainers.SortedKeyList([], key=lambda i: i["name"].lower())
        async with aiohttp.ClientSession() as session:
            for url in [
                "https://5e.tools/data/spells/spells-ai.json",
                "https://5e.tools/data/spells/spells-ggr.json",
                "https://5e.tools/data/spells/spells-llk.json",
                "https://5e.tools/data/spells/spells-phb.json",
                "https://5e.tools/data/spells/spells-scag.json",
                "https://5e.tools/data/spells/spells-stream.json",
                "https://5e.tools/data/spells/spells-ua-ar.json",
                "https://5e.tools/data/spells/spells-ua-mm.json",
                "https://5e.tools/data/spells/spells-ua-ss.json",
                "https://5e.tools/data/spells/spells-ua-tobm.json",
                "https://5e.tools/data/spells/spells-xge.json"
            ]:
                async with session.get(url) as response:
                    j = await response.json()
                    for spell in j["spell"]:
                        self._dnddata.add(spell)

    def _parse_spell(self, spell: dict) -> str:
        string = f'✨ [b]{spell["name"]}[/b]\n'
        if "source" in spell:
            string += f'[i]{spell["source"]}, page {spell["page"]}[/i]\n'
        string += "\n"
        if spell["level"] == 0:
            string += f'[b]Cantrip[/b] {spell["school"]}\n'
        else:
            string += f'[b]{ordinalformat(spell["level"])}[/b] level {spell["school"]}\n'
        if "time" in spell:
            for time in spell["time"]:
                string += f'Cast time: ⌛️ [b]{time["number"]} {time["unit"]}[/b]\n'
        if "range" in spell:
            if spell["range"]["distance"]["type"] == "touch":
                string += "Range: 👉 [b]Touch[/b]\n"
            elif spell["range"]["distance"]["type"] == "self":
                string += "Range: 👤 [b]Self[/b]\n"
            else:
                string += f'Range: 🏹 [b]{spell["range"]["distance"]["amount"]} {spell["range"]["distance"]["type"]}[/b] ({spell["range"]["type"]})\n'
        if "components" in spell:
            string += f'Components: '
            if spell["components"].get("v", False):
                string += "👄 [b]Verbal[/b] | "
            if spell["components"].get("s", False):
                string += "🤙 [b]Somatic[/b] | "
            if spell["components"].get("r", False):
                # TODO: wtf is this
                string += "❓ [b]R...?[/b] | "
            if spell["components"].get("m", False):
                if "text" in spell["components"]["m"]:
                    string += f'💎 [b]Material[/b] ([i]{spell["components"]["m"]["text"]}[/i]) | '
                else:
                    string += f'💎 [b]Material[/b] ([i]{spell["components"]["m"]}[/i]) | '
                string = string.rstrip(" ").rstrip("|")
            string += "\n"
        string += "\n"
        if "duration" in spell:
            for duration in spell["duration"]:
                if duration["type"] == "timed":
                    string += f'Duration: 🕒 [b]{duration["duration"]["amount"]} {duration["duration"]["type"]}[/b]'
                elif duration["type"] == "instant":
                    string += 'Duration: ☁️ [b]Instantaneous[/b]'
                elif duration["type"] == "special":
                    string += 'Duration: ⭐️ [b]Special[/b]'
                elif duration["type"] == "permanent":
                    string += f"Duration: ♾ [b]Permanent[/b] (ends on {andformat(duration['ends'], final=' or ')})"
                else:
                    string += f'Duration: ⚠️[b]UNKNOWN[/b]'
                if duration.get("concentration", False):
                    string += " (requires concentration)"
                string += "\n"
        if "meta" in spell:
            if spell["meta"].get("ritual", False):
                string += "🔮 Can be casted as ritual\n"
        string += "\n"
        for entry in spell.get("entries", []):
            string += parse_5etools_entry(entry)
            string += "\n\n"
        for entry in spell.get("entriesHigherLevel", []):
            string += parse_5etools_entry(entry)
            string += "\n\n"
        return string

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self._dnddata is None:
            await data.reply("⚠️ Il database degli oggetti di D&D non è ancora stato scaricato.")
            return
        search = args.joined().lower()
        result = self._dnddata[self._dnddata.bisect_key_left(search)]
        await data.reply(self._parse_spell(result))



import typing
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData


class DebugKeyboardCommand(Command):
    name: str = "debug_keyboard"

    description: str = "Invia una tastiera di prova."

    async def _callback(self, data: CommandData):
        await data.reply("OK.")

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        await data.keyboard("This is a keyboard.", {
            "✅ OK": self._callback
        })



import typing
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData


class DebugErrorCommand(Command):
    name: str = "debug_error"

    description: str = "Causa un'eccezione nel bot."

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        raise Exception("debug_error command was called")



"""Some Royalprints that can be used with the Royalnet Flask server."""

from . import home
from . import wikiview
from . import tglogin
from . import docs
from . import wikiedit
from . import mcstatus
from . import diarioview
from . import profile
from . import login
from . import newaccount

rp_home = home.rp
rp_wikiview = wikiview.rp
rp_tglogin = tglogin.rp
rp_docs = docs.rp
rp_wikiedit = wikiedit.rp
rp_mcstatus = mcstatus.rp
rp_diarioview = diarioview.rp
rp_profile = profile.rp
rp_login = login.rp
rp_newaccount = newaccount.rp

__all__ = ["rp_home", "rp_wikiview", "rp_tglogin", "rp_docs", "rp_wikiedit", "rp_mcstatus", "rp_diarioview",
           "rp_profile", "rp_login", "rp_newaccount"]



def parse_5etools_entry(entry) -> str:
    if isinstance(entry, str):
        return entry
    elif isinstance(entry, dict):
        string = ""
        if entry["type"] == "entries":
            string += f'[b]{entry.get("name", "")}[/b]\n'
            for subentry in entry["entries"]:
                string += parse_5etools_entry(subentry)
                string += "\n\n"
        elif entry["type"] == "table":
            string += "[i][table hidden][/i]"
            # for label in entry["colLabels"]:
            #     string += f"| {label} "
            # string += "|"
            # for row in entry["rows"]:
            #     for column in row:
            #         string += f"| {self._parse_entry(column)} "
            #     string += "|\n"
        elif entry["type"] == "cell":
            return parse_5etools_entry(entry["entry"])
        elif entry["type"] == "list":
            string = ""
            for item in entry["items"]:
                string += f"- {parse_5etools_entry(item)}\n"
            string.rstrip("\n")
        else:
            string += "[i]⚠️ [unknown type][/i]"
    else:
        return "[/i]⚠️ [unknown data][/i]"
    return string



import typing
import discord
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...utils import NetworkHandler
from ...network import Request, ResponseSuccess
from ...error import NoneFoundError
if typing.TYPE_CHECKING:
    from ...bots import DiscordBot


class SummonNH(NetworkHandler):
    message_type = "music_summon"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        """Handle a summon Royalnet request.
         That is, join a voice channel, or move to a different one if that is not possible."""
        channel = bot.client.find_channel_by_name(data["channel_name"])
        if not isinstance(channel, discord.VoiceChannel):
            raise NoneFoundError("Channel is not a voice channel")
        bot.loop.create_task(bot.client.vc_connect_or_move(channel))
        return ResponseSuccess()


class SummonCommand(Command):
    name: str = "summon"

    description: str = "Evoca il bot in un canale vocale."

    syntax: str = "[nomecanale]"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler(SummonNH.message_type, SummonNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self.interface.name == "discord":
            bot = self.interface.bot.client
            message: discord.Message = data.message
            channel_name: str = args.optional(0)
            if channel_name:
                guild: typing.Optional[discord.Guild] = message.guild
                if guild is not None:
                    channels: typing.List[discord.abc.GuildChannel] = guild.channels
                else:
                    channels = bot.get_all_channels()
                matching_channels: typing.List[discord.VoiceChannel] = []
                for channel in channels:
                    if isinstance(channel, discord.VoiceChannel):
                        if channel.name == channel_name:
                            matching_channels.append(channel)
                if len(matching_channels) == 0:
                    await data.reply("⚠️ Non esiste alcun canale vocale con il nome specificato.")
                    return
                elif len(matching_channels) > 1:
                    await data.reply("⚠️ Esiste più di un canale vocale con il nome specificato.")
                    return
                channel = matching_channels[0]
            else:
                author: discord.Member = message.author
                try:
                    voice: typing.Optional[discord.VoiceState] = author.voice
                except AttributeError:
                    await data.reply("⚠️ Non puoi evocare il bot da una chat privata!")
                    return
                if voice is None:
                    await data.reply("⚠️ Non sei connesso a nessun canale vocale!")
                    return
                channel = voice.channel
            await bot.vc_connect_or_move(channel)
            await data.reply(f"✅ Mi sono connesso in [c]#{channel.name}[/c].")
        else:
            channel_name: str = args[0].lstrip("#")
            await self.interface.net_request(Request(SummonNH.message_type, {"channel_name": channel_name}), "discord")
            await data.reply(f"✅ Mi sono connesso in [c]#{channel_name}[/c].")



import telegram
import typing
from sqlalchemy import Column, \
                       Integer, \
                       DateTime, \
                       String, \
                       Text, \
                       ForeignKey, \
                       BigInteger
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
from .royals import Royal
if typing.TYPE_CHECKING:
    from .mmdecisions import MMDecision
    from .mmresponse import MMResponse


class MMEvent:
    __tablename__ = "mmevents"

    @declared_attr
    def creator_id(self):
        return Column(Integer, ForeignKey("royals.uid"), nullable=False)

    @declared_attr
    def creator(self):
        return relationship("Royal", backref="mmevents_created")

    @declared_attr
    def mmid(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def datetime(self):
        return Column(DateTime, nullable=False)

    @declared_attr
    def title(self):
        return Column(String, nullable=False)

    @declared_attr
    def description(self):
        return Column(Text, nullable=False, default="")

    @declared_attr
    def state(self):
        # Valid states are WAITING, DECISION, READY_CHECK, STARTED
        return Column(String, nullable=False, default="WAITING")

    @declared_attr
    def message_id(self):
        return Column(BigInteger)

    def __repr__(self):
        return f"<MMEvent {self.mmid}: {self.title}>"




import math
import random
import typing
from collections import namedtuple
from .ytdldiscord import YtdlDiscord
from .fileaudiosource import FileAudioSource


class PlayMode:
    """The base class for a PlayMode, such as :py:class:`royalnet.audio.Playlist`. Inherit from this class if you want to create a custom PlayMode."""

    def __init__(self):
        """Create a new PlayMode and initialize the generator inside."""
        self.now_playing: typing.Optional[YtdlDiscord] = None
        self.generator: typing.AsyncGenerator = self._generate_generator()

    async def next(self) -> typing.Optional[FileAudioSource]:
        """Get the next :py:class:`royalnet.audio.FileAudioSource` from the list and advance it.

        Returns:
            The next :py:class:`royalnet.audio.FileAudioSource`."""
        return await self.generator.__anext__()

    def videos_left(self) -> typing.Union[int, float]:
        """Return the number of videos left in the PlayMode.

        Returns:
            Usually a :py:class:`int`, but may return also :py:obj:`math.inf` if the PlayMode is infinite."""
        raise NotImplementedError()

    async def _generate_generator(self):
        """Factory function for an async generator that changes the ``now_playing`` property either to a :py:class:`royalnet.audio.FileAudioSource` or to ``None``, then yields the value it changed it to.

        Yields:
            The :py:class:`royalnet.audio.FileAudioSource` to be played next."""
        raise NotImplementedError()
        # This is needed to make the coroutine an async generator
        # noinspection PyUnreachableCode
        yield NotImplemented

    def add(self, item: YtdlDiscord) -> None:
        """Add a new :py:class:`royalnet.audio.YtdlDiscord` to the PlayMode.

        Args:
            item: The item to add to the PlayMode."""
        raise NotImplementedError()

    def delete(self) -> None:
        """Delete all :py:class:`royalnet.audio.YtdlDiscord` contained inside this PlayMode."""
        raise NotImplementedError()

    def queue_preview(self) -> typing.List[YtdlDiscord]:
        """Display all the videos in the PlayMode as a list, if possible.

        To be used with ``queue`` commands, for example.

        Raises:
            NotImplementedError: If a preview can't be generated.

        Returns:
            A list of videos contained in the queue."""
        raise NotImplementedError()


class Playlist(PlayMode):
    """A video list. :py:class:`royalnet.audio.YtdlDiscord` played are removed from the list."""

    def __init__(self, starting_list: typing.List[YtdlDiscord] = None):
        """Create a new Playlist.

        Args:
            starting_list: A list of items with which the Playlist will be created."""
        super().__init__()
        if starting_list is None:
            starting_list = []
        self.list: typing.List[YtdlDiscord] = starting_list

    def videos_left(self) -> typing.Union[int, float]:
        return len(self.list)

    async def _generate_generator(self):
        while True:
            try:
                next_video = self.list.pop(0)
            except IndexError:
                self.now_playing = None
                yield None
            else:
                self.now_playing = next_video
                yield next_video.spawn_audiosource()
            if self.now_playing is not None:
                self.now_playing.delete()

    def add(self, item) -> None:
        self.list.append(item)

    def delete(self) -> None:
        if self.now_playing is not None:
            self.now_playing.delete()
        while self.list:
            self.list.pop(0).delete()

    def queue_preview(self) -> typing.List[YtdlDiscord]:
        return self.list


class Pool(PlayMode):
    """A random pool. :py:class:`royalnet.audio.YtdlDiscord` are selected in random order and are not repeated until every song has been played at least once."""

    def __init__(self, starting_pool: typing.List[YtdlDiscord] = None):
        """Create a new Pool.

        Args:
            starting_pool: A list of items the Pool will be created from."""
        super().__init__()
        if starting_pool is None:
            starting_pool = []
        self.pool: typing.List[YtdlDiscord] = starting_pool
        self._pool_copy: typing.List[YtdlDiscord] = []

    def videos_left(self) -> typing.Union[int, float]:
        return math.inf

    async def _generate_generator(self):
        while True:
            if not self.pool:
                self.now_playing = None
                yield None
                continue
            self._pool_copy = self.pool.copy()
            random.shuffle(self._pool_copy)
            while self._pool_copy:
                next_video = self._pool_copy.pop(0)
                self.now_playing = next_video
                yield next_video.spawn_audiosource()

    def add(self, item) -> None:
        self.pool.append(item)
        self._pool_copy.append(item)
        random.shuffle(self._pool_copy)

    def delete(self) -> None:
        for item in self.pool:
            item.delete()
        self.pool = None
        self._pool_copy = None

    def queue_preview(self) -> typing.List[YtdlDiscord]:
        preview_pool = self.pool.copy()
        random.shuffle(preview_pool)
        return preview_pool


class Layers(PlayMode):
    """A playmode for playing a single song with multiple layers."""

    Layer = namedtuple("Layer", ["dfile", "source"])

    def __init__(self, starting_layers: typing.List[YtdlDiscord] = None):
        super().__init__()
        if starting_layers is None:
            starting_layers = []
        self.layers = []
        for item in starting_layers:
            self.add(item)

    def videos_left(self) -> typing.Union[int, float]:
        return 1 if len(self.layers) > 0 else 0

    async def _generate_generator(self):
        current_layer = None
        current_source = None
        while True:
            if len(self.layers) == 0:
                yield None
                continue
            if self.now_playing is None:
                self.now_playing = self.layers[0].dfile
                current_source = self.layers[0].source
                current_layer = 0
                yield current_source
                continue
            if current_source.file.closed:
                self.now_playing = None
                self.layers = []
                current_layer = None
                current_source = None
                yield None
                continue
            current_layer += 1
            current_position = current_source.file.tell()
            if current_layer >= len(self.layers):
                self.now_playing = self.layers[0].dfile
                current_source = self.layers[0].source
                current_source.file.seek(current_position)
                current_layer = 0
                yield current_source
                continue
            self.now_playing = self.layers[current_layer].dfile
            current_source = self.layers[current_layer].source
            current_source.file.seek(current_position)
            yield current_source

    def add(self, item) -> None:
        self.layers.append(self.Layer(dfile=item, source=item.spawn_audiosource()))

    def delete(self) -> None:
        for item in self.layers:
            item.dfile.delete()
        self.layers = None

    def queue_preview(self) -> typing.List[YtdlDiscord]:
        return [layer.dfile for layer in self.layers]



import typing
import re
import ffmpeg
import os
from .ytdlinfo import YtdlInfo
from .ytdlfile import YtdlFile
from .fileaudiosource import FileAudioSource


class YtdlDiscord:
    def __init__(self, ytdl_file: YtdlFile):
        self.ytdl_file: YtdlFile = ytdl_file
        self.pcm_filename: typing.Optional[str] = None
        self._fas_spawned: typing.List[FileAudioSource] = []

    def pcm_available(self):
        return self.pcm_filename is not None and os.path.exists(self.pcm_filename)

    def convert_to_pcm(self) -> None:
        if not self.ytdl_file.is_downloaded():
            raise FileNotFoundError("File hasn't been downloaded yet")
        destination_filename = re.sub(r"\.[^.]+$", ".pcm", self.ytdl_file.filename)
        (
            ffmpeg.input(self.ytdl_file.filename)
                  .output(destination_filename, format="s16le", ac=2, ar="48000")
                  .overwrite_output()
                  .run(quiet=True)
        )
        self.pcm_filename = destination_filename

    def ready_up(self):
        if not self.ytdl_file.has_info():
            self.ytdl_file.update_info()
        if not self.ytdl_file.is_downloaded():
            self.ytdl_file.download_file()
        if not self.pcm_available():
            self.convert_to_pcm()

    def spawn_audiosource(self) -> FileAudioSource:
        if not self.pcm_available():
            raise FileNotFoundError("File hasn't been converted to PCM yet")
        stream = open(self.pcm_filename, "rb")
        source = FileAudioSource(stream)
        # FIXME: it's a intentional memory leak
        self._fas_spawned.append(source)
        return source

    def delete(self) -> None:
        if self.pcm_available():
            for source in self._fas_spawned:
                if not source.file.closed:
                    source.file.close()
            os.remove(self.pcm_filename)
            self.pcm_filename = None
        self.ytdl_file.delete()

    @classmethod
    def create_from_url(cls, url, **ytdl_args) -> typing.List["YtdlDiscord"]:
        files = YtdlFile.download_from_url(url, **ytdl_args)
        dfiles = []
        for file in files:
            dfile = YtdlDiscord(file)
            dfiles.append(dfile)
        return dfiles

    @classmethod
    def create_and_ready_from_url(cls, url, **ytdl_args) -> typing.List["YtdlDiscord"]:
        dfiles = cls.create_from_url(url, **ytdl_args)
        for dfile in dfiles:
            dfile.ready_up()
        return dfiles

    @property
    def info(self) -> typing.Optional[YtdlInfo]:
        return self.ytdl_file.info



"""The production Royalnet available at ryg.steffo.eu ."""

import os
from royalnet.web import create_app
from royalnet.web.royalprints import *


class TestConfig:
    DB_PATH = os.environ["DB_PATH"]
    TG_AK = os.environ["TG_AK"]
    SITE_NAME = "Royalnet"


app = create_app(TestConfig, [rp_home, rp_wikiview, rp_tglogin, rp_docs, rp_wikiedit, rp_mcstatus, rp_diarioview,
                              rp_profile, rp_login])


if __name__ == "__main__":
    app.run()



import typing
import discord
import asyncio
import datetime
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...utils import NetworkHandler, asyncify
from ...network import Request, ResponseSuccess
from ...error import *
from ...audio import YtdlDiscord
from ...audio.playmodes import Playlist
if typing.TYPE_CHECKING:
    from ...bots import DiscordBot


class ZawarudoNH(NetworkHandler):
    message_type = "music_zawarudo"

    ytdl_args = {
        "format": "bestaudio",
        "outtmpl": f"./downloads/%(title)s.%(ext)s"
    }

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Ensure the guild has a PlayMode before adding the file to it
        if not bot.music_data.get(guild):
            # TODO: change Exception
            raise Exception("No music_data for this guild")
        # Create url
        ytdl_args = {
            "format": "bestaudio",
            "outtmpl": f"./downloads/{datetime.datetime.now().timestamp()}_%(title)s.%(ext)s"
        }
        # Start downloading
        zw_start: typing.List[YtdlDiscord] = await asyncify(YtdlDiscord.create_and_ready_from_url,
                                                            "https://scaleway.steffo.eu/jojo/zawarudo_intro.mp3",
                                                            **ytdl_args)
        zw_end: typing.List[YtdlDiscord] = await asyncify(YtdlDiscord.create_and_ready_from_url,
                                                          "https://scaleway.steffo.eu/jojo/zawarudo_outro.mp3",
                                                          **ytdl_args)
        old_playlist = bot.music_data[guild]
        bot.music_data[guild] = Playlist()
        # Get voice client
        vc: discord.VoiceClient = bot.client.find_voice_client_by_guild(guild)
        channel: discord.VoiceChannel = vc.channel
        affected: typing.List[typing.Union[discord.User, discord.Member]] = channel.members
        await bot.add_to_music_data(zw_start, guild)
        for member in affected:
            if member.bot:
                continue
            await member.edit(mute=True)
        await asyncio.sleep(data["time"])
        await bot.add_to_music_data(zw_end, guild)
        for member in affected:
            member: typing.Union[discord.User, discord.Member]
            if member.bot:
                continue
            await member.edit(mute=False)
        bot.music_data[guild] = old_playlist
        await bot.advance_music_data(guild)
        return ResponseSuccess()


class ZawarudoCommand(Command):
    name: str = "zawarudo"

    description: str = "Ferma il tempo!"

    syntax = "[ [guild] ] [1-9]"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler(ZawarudoNH.message_type, ZawarudoNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild_name, time = args.match(r"(?:\[(.+)])?\s*(.+)?")
        if time is None:
            time = 5
        else:
            time = int(time)
        if time < 1:
            raise InvalidInputError("The World can't stop time for less than a second.")
        if time > 10:
            raise InvalidInputError("The World can stop time only for 10 seconds.")
        await data.reply(f"🕒 ZA WARUDO! TOKI WO TOMARE!")
        await self.interface.net_request(Request(ZawarudoNH.message_type, {"time": time, "guild_name": guild_name}), "discord")



import typing
import telegram
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData


class CiaoruoziCommand(Command):
    name: str = "ciaoruozi"

    description: str = "Saluta Ruozi, un leggendario essere che una volta era in Royal Games."

    syntax: str = ""

    require_alchemy_tables: typing.Set = set()

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self.interface.name == "telegram":
            update: telegram.Update = data.update
            user: telegram.User = update.effective_user
            if user.id == 112437036:
                await data.reply("👋 Ciao me!")
                return
        await data.reply("👋 Ciao Ruozi!")



from .royals import Royal
from .telegram import Telegram
from .diario import Diario
from .aliases import Alias
from .activekvgroups import ActiveKvGroup
from .keyvalues import Keyvalue
from .keygroups import Keygroup
from .discord import Discord
from .wikipages import WikiPage
from .wikirevisions import WikiRevision
from .medals import Medal
from .medalawards import MedalAward
from .bios import Bio
from .reminders import Reminder
from .triviascores import TriviaScore
from .mmdecisions import MMDecision
from .mmevents import MMEvent
from .mmresponse import MMResponse

__all__ = ["Royal", "Telegram", "Diario", "Alias", "ActiveKvGroup", "Keyvalue", "Keygroup", "Discord", "WikiPage",
           "WikiRevision", "Medal", "MedalAward", "Bio", "Reminder", "TriviaScore", "MMDecision", "MMEvent",
           "MMResponse"]



import typing
import asyncio
import aiohttp
import random
import uuid
import html
from ..command import Command
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ..commandinterface import CommandInterface
from ...utils import asyncify
from ...error import *
from ...database.tables import TriviaScore


class TriviaCommand(Command):
    name: str = "trivia"

    description: str = "Manda una domanda dell'OpenTDB in chat."

    require_alchemy_tables = {TriviaScore}

    _letter_emojis = ["🇦", "🇧", "🇨", "🇩"]

    _correct_emoji = "✅"

    _wrong_emoji = "❌"

    _answer_time = 18

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        self._answerers: typing.Dict[uuid.UUID, typing.Dict[..., bool]] = {}

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        # Fetch the question
        async with aiohttp.ClientSession() as session:
            async with session.get("https://opentdb.com/api.php?amount=1") as response:
                j = await response.json()
        # Parse the question
        if j["response_code"] != 0:
            raise ExternalError(f"OpenTDB returned {j['response_code']} response_code")
        question = j["results"][0]
        text = f'❓ [b]{question["category"]} - {question["difficulty"].capitalize()}[/b]\n' \
               f'{html.unescape(question["question"])}'
        # Prepare answers
        correct_answer: str = question["correct_answer"]
        wrong_answers: typing.List[str] = question["incorrect_answers"]
        answers: typing.List[str] = [correct_answer, *wrong_answers]
        if question["type"] == "multiple":
            random.shuffle(answers)
        elif question["type"] == "boolean":
            answers.sort(key=lambda a: a)
            answers.reverse()
        else:
            raise NotImplementedError("Unknown question type")
        # Find the correct index
        for index, answer in enumerate(answers):
            if answer == correct_answer:
                correct_index = index
                break
        else:
            raise ValueError("correct_index not found")
        # Add emojis
        for index, answer in enumerate(answers):
            answers[index] = f"{self._letter_emojis[index]} {html.unescape(answers[index])}"
        # Create the question id
        question_id = uuid.uuid4()
        self._answerers[question_id] = {}

        # Create the correct and wrong functions
        async def correct(data: CommandData):
            answerer_ = await data.get_author(error_if_none=True)
            try:
                self._answerers[question_id][answerer_] = True
            except KeyError:
                raise KeyboardExpiredError("Question time ran out.")
            return "🆗 Hai risposto alla domanda. Ora aspetta un attimo per i risultati!"

        async def wrong(data: CommandData):
            answerer_ = await data.get_author(error_if_none=True)
            try:
                self._answerers[question_id][answerer_] = False
            except KeyError:
                raise KeyboardExpiredError("Question time ran out.")
            return "🆗 Hai risposto alla domanda. Ora aspetta un attimo per i risultati!"

        # Add question
        keyboard = {}
        for index, answer in enumerate(answers):
            if index == correct_index:
                keyboard[answer] = correct
            else:
                keyboard[answer] = wrong
        await data.keyboard(text, keyboard)
        await asyncio.sleep(self._answer_time)
        results = f"❗️ Tempo scaduto!\n" \
                  f"La risposta corretta era [b]{answers[correct_index]}[/b]!\n\n"
        for answerer in self._answerers[question_id]:
            if answerer.trivia_score is None:
                ts = self.interface.alchemy.TriviaScore(royal=answerer)
                self.interface.session.add(ts)
                self.interface.session.commit()
            if self._answerers[question_id][answerer]:
                results += self._correct_emoji
                answerer.trivia_score.correct_answers += 1
            else:
                results += self._wrong_emoji
                answerer.trivia_score.wrong_answers += 1
            results += f" {answerer} ({answerer.trivia_score.correct_answers}/{answerer.trivia_score.total_answers})\n"
        await data.reply(results)
        del self._answerers[question_id]
        await asyncify(self.interface.session.commit)



import typing
import asyncio
if typing.TYPE_CHECKING:
    from ..database import Alchemy
    from ..bots import GenericBot


class CommandInterface:
    name: str = NotImplemented
    prefix: str = NotImplemented
    alchemy: "Alchemy" = NotImplemented
    bot: "GenericBot" = NotImplemented
    loop: asyncio.AbstractEventLoop = NotImplemented

    def __init__(self):
        self.session = self.alchemy.Session()

    def register_net_handler(self, message_type: str, network_handler: typing.Callable):
        """Register a new handler for messages received through Royalnet."""
        raise NotImplementedError()

    def unregister_net_handler(self, message_type: str):
        """Remove a Royalnet handler."""
        raise NotImplementedError()

    async def net_request(self, message, destination: str) -> dict:
        """Send data through a :py:class:`royalnet.network.RoyalnetLink` and wait for a
        :py:class:`royalnet.network.Reply`.

        Parameters:
            message: The data to be sent. Must be :py:mod:`pickle`-able.
            destination: The destination of the request, either in UUID format or node name."""
        raise NotImplementedError()

    def register_keyboard_key(self, key_name: str, callback: typing.Callable):
        raise NotImplementedError()

    def unregister_keyboard_key(self, key_name: str):
        raise NotImplementedError()



import typing
from ..error import UnsupportedError
from .commandinterface import CommandInterface
from .commandargs import CommandArgs
from .commanddata import CommandData


class Command:
    name: str = NotImplemented
    """The main name of the command.
     To have ``/example`` on Telegram, the name should be ``example``."""

    description: str = NotImplemented
    """A small description of the command, to be displayed when the command is being autocompleted."""

    syntax: str = ""
    """The syntax of the command, to be displayed when a :py:exc:`royalnet.error.InvalidInputError` is raised,
     in the format ``(required_arg) [optional_arg]``."""

    require_alchemy_tables: typing.Set = set()
    """A set of :py:class:`royalnet.database` tables that must exist for this command to work."""

    def __init__(self, interface: CommandInterface):
        self.interface = interface

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        raise UnsupportedError(f"Command {self.name} can't be called on {self.interface.name}.")



"""The profile page :py:class:`royalnet.web.Royalprint` for Royalnet members."""

import flask as f
import os
from ...royalprint import Royalprint
from ...shortcuts import error
from ....database.tables import *
from ....utils.wikirender import prepare_page_markdown, RenderError


# Maybe some of these tables are optional...
tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("profile", __name__, url_prefix="/profile", template_folder=tmpl_dir,
                required_tables={Royal, ActiveKvGroup, Alias, Diario, Discord, Keygroup, Keyvalue, Telegram, WikiPage,
                                 WikiRevision, Bio, TriviaScore})


@rp.route("/")
def profile_index():
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    royals = alchemy_session.query(alchemy.Royal).order_by(alchemy.Royal.username).all()
    return f.render_template("profile_index.html", royals=royals)


@rp.route("/<username>")
def profile_page(username):
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    royal = alchemy_session.query(alchemy.Royal).filter_by(username=username).one_or_none()
    if royal is None:
        return error(404, "Non esiste nessun utente con l'username richiesto.")
    if royal.bio is not None and royal.bio.contents != "":
        try:
            parsed_bio = f.Markup(prepare_page_markdown(royal.bio.contents))
        except RenderError as e:
            return error(500, f"Il profilo non può essere visualizzato a causa di un errore nella bio: {str(e)}")
    else:
        parsed_bio = None
    return f.render_template("profile_page.html", royal=royal, parsed_bio=parsed_bio)


@rp.route("/<username>/editbio", methods=["GET", "POST"])
def profile_editbio(username):
    if "royal" not in f.session:
        return error(403, "Devi aver effettuato il login per modificare una bio.")
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    royal = alchemy_session.query(alchemy.Royal).filter_by(username=username).one_or_none()
    if not (f.session["royal"]["uid"] == royal.uid or f.session["royal"]["role"] == "Admin"):
        return error(403, "Non sei autorizzato a modificare questa pagina bio.")

    if f.request.method == "GET":
        return f.render_template("profile_editbio.html", royal=royal)

    elif f.request.method == "POST":
        fd = f.request.form
        if royal.bio is None:
            bio = alchemy.Bio(royal=royal, contents=fd.get("contents", ""))
            alchemy_session.add(bio)
        else:
            royal.bio.contents = fd.get("contents", "")
        alchemy_session.commit()
        return f.redirect(f.url_for(".profile_page", username=username))



"""Miscellaneous useful functions and classes."""

from .asyncify import asyncify
from .escaping import telegram_escape, discord_escape
from .safeformat import safeformat
from .classdictjanitor import cdj
from .sleepuntil import sleep_until
from .networkhandler import NetworkHandler
from .formatters import andformat, plusformat, fileformat, ytdldateformat, numberemojiformat, splitstring, ordinalformat
from .parse5etoolsentry import parse_5etools_entry

__all__ = ["asyncify", "safeformat", "cdj", "sleep_until", "plusformat",
           "NetworkHandler", "andformat", "plusformat", "fileformat", "ytdldateformat", "numberemojiformat",
           "telegram_escape", "discord_escape", "splitstring", "parse_5etools_entry", "ordinalformat"]



import typing
import pickle
from ..command import Command
from ..commandinterface import CommandInterface
from ..commandargs import CommandArgs
from ..commanddata import CommandData
from ...utils import NetworkHandler
from ...network import Request, ResponseSuccess
from ...error import *
from ...audio.playmodes import Playlist, Pool, Layers
if typing.TYPE_CHECKING:
    from ...bots import DiscordBot


class PlaymodeNH(NetworkHandler):
    message_type = "music_playmode"

    @classmethod
    async def discord(cls, bot: "DiscordBot", data: dict):
        """Handle a playmode Royalnet request. That is, change current PlayMode."""
        # Find the matching guild
        if data["guild_name"]:
            guild = bot.client.find_guild(data["guild_name"])
        else:
            if len(bot.music_data) == 0:
                raise NoneFoundError("No voice clients active")
            if len(bot.music_data) > 1:
                raise TooManyFoundError("Multiple guilds found")
            guild = list(bot.music_data)[0]
        # Delete the previous PlayMode, if it exists
        if bot.music_data[guild] is not None:
            bot.music_data[guild].delete()
        # Create the new PlayMode
        if data["mode_name"] == "playlist":
            bot.music_data[guild] = Playlist()
        elif data["mode_name"] == "pool":
            bot.music_data[guild] = Pool()
        elif data["mode_name"] == "layers":
            bot.music_data[guild] = Layers()
        else:
            raise ValueError("No such PlayMode")
        return ResponseSuccess()


class PlaymodeCommand(Command):
    name: str = "playmode"

    description: str = "Cambia modalità di riproduzione per la chat vocale."

    syntax = "[ [guild] ] (mode)"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        interface.register_net_handler(PlaymodeNH.message_type, PlaymodeNH)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild_name, mode_name = args.match(r"(?:\[(.+)])?\s*(\S+)\s*")
        await self.interface.net_request(Request(PlaymodeNH.message_type, {"mode_name": mode_name,
                                                                           "guild_name": guild_name}),
                                         "discord")
        await data.reply(f"🔃 Impostata la modalità di riproduzione a: [c]{mode_name}[/c].")



"""Commands that can be used in bots.

These probably won't suit your needs, as they are tailored for the bots of the Royal Games gaming community, but they
 may be useful to develop new ones."""

from .ciaoruozi import CiaoruoziCommand
from .color import ColorCommand
from .cv import CvCommand
from .diario import DiarioCommand
from .mp3 import Mp3Command
from .pause import PauseCommand
from .ping import PingCommand
from .play import PlayCommand
from .playmode import PlaymodeCommand
from .queue import QueueCommand
from .rage import RageCommand
from .reminder import ReminderCommand
from .ship import ShipCommand
from .skip import SkipCommand
from .smecds import SmecdsCommand
from .summon import SummonCommand
from .videochannel import VideochannelCommand
from .dnditem import DnditemCommand
from .dndspell import DndspellCommand
from .trivia import TriviaCommand
from .mm import MmCommand
from .zawarudo import ZawarudoCommand

__all__ = [
    "CiaoruoziCommand",
    "ColorCommand",
    "CvCommand",
    "DiarioCommand",
    "Mp3Command",
    "PauseCommand",
    "PingCommand",
    "PlayCommand",
    "PlaymodeCommand",
    "QueueCommand",
    "RageCommand",
    "ReminderCommand",
    "ShipCommand",
    "SkipCommand",
    "SmecdsCommand",
    "SummonCommand",
    "VideochannelCommand",
    "DnditemCommand",
    "DndspellCommand",
    "TriviaCommand",
    "MmCommand",
    "ZawarudoCommand"
]



"""A Royalnet password-based login :py:class:`royalnet.web.Royalprint`."""
import flask as f
import os
import datetime
import bcrypt
from ...royalprint import Royalprint
from ...shortcuts import error
from ....database.tables import Royal


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("login", __name__, url_prefix="/login/password", required_tables={Royal},
                template_folder=tmpl_dir)


@rp.route("/")
def login_index():
    f.session.pop("royal", None)
    return f.render_template("login_index.html")


@rp.route("/done", methods=["POST"])
def login_done():
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    fd = f.request.form
    if "username" not in fd:
        return error(400, "Nessun username inserito.")
    royal_user = alchemy_session.query(alchemy.Royal).filter_by(username=fd["username"]).one_or_none()
    if royal_user is None:
        return error(404, "L'username inserito non corrisponde a nessun account registrato.")
    if "password" not in fd:
        return error(400, "Nessuna password inserita.")
    if not bcrypt.checkpw(bytes(fd["password"], encoding="utf8"), royal_user.password):
        return error(400, "La password inserita non è valida.")
    f.session["royal"] = {
        "uid": royal_user.uid,
        "username": royal_user.username,
        "avatar": royal_user.avatar,
        "role": royal_user.role
    }
    f.session["login_date"] = datetime.datetime.now()
    return f.render_template("login_success.html")
